"""API for Google Photos OAuth.

Callers subclass this to provide an asyncio implementation that refreshes
authentication tokens.
"""

from http import HTTPStatus
from abc import ABC, abstractmethod
import logging
from typing import Any, TypeVar, Type
from mashumaro.mixins.json import DataClassJSONMixin

import aiohttp

from aiohttp.client_exceptions import ClientError

from .exceptions import (
    ApiForbiddenException,
    ApiException,
    AuthException,
)
from .const import LIBRARY_API_URL
from .model import ErrorResponse, Error

__all__ = ["AbstractAuth"]

_LOGGER = logging.getLogger(__name__)


AUTHORIZATION_HEADER = "Authorization"
ERROR = "error"
STATUS = "status"
MESSAGE = "message"

_T = TypeVar("_T", bound=DataClassJSONMixin)


class AbstractAuth(ABC):
    """Base class for Google Photos authentication library.

    Provides an asyncio interface around the blocking client library.
    """

    def __init__(self, websession: aiohttp.ClientSession, host: str | None = None):
        """Initialize the auth."""
        self._websession = websession
        self._host = host or LIBRARY_API_URL

    @abstractmethod
    async def async_get_access_token(self) -> str:
        """Return a valid access token."""

    async def request(
        self,
        method: str,
        url: str,
        headers: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> aiohttp.ClientResponse:
        """Make a request."""
        try:
            access_token = await self.async_get_access_token()
        except ClientError as err:
            raise AuthException(f"Access token failure: {err}") from err
        if headers is None:
            headers = {}
        if AUTHORIZATION_HEADER not in headers:
            headers[AUTHORIZATION_HEADER] = f"Bearer {access_token}"
        if not (url.startswith("http://") or url.startswith("https://")):
            url = f"{self._host}/{url}"
        _LOGGER.debug("request[%s]=%s %s", method, url, kwargs.get("params"))
        if method != "get" and "json" in kwargs:
            _LOGGER.debug("request[post json]=%s", kwargs["json"])
        return await self._websession.request(method, url, **kwargs, headers=headers)

    async def get(self, url: str, **kwargs: Any) -> aiohttp.ClientResponse:
        """Make a get request."""
        try:
            resp = await self.request("get", url, **kwargs)
        except ClientError as err:
            raise ApiException(f"Error connecting to API: {err}") from err
        return await AbstractAuth._raise_for_status(resp)

    async def get_json(
        self,
        url: str,
        data_cls: Type[_T],
        **kwargs: Any,
    ) -> _T:
        """Make a get request and return json response."""
        resp = await self.get(url, **kwargs)
        try:
            result = await resp.text()
        except ClientError as err:
            raise ApiException("Server returned malformed response") from err
        _LOGGER.debug("response=%s", result)
        try:
            return data_cls.from_json(result)
        except (LookupError, ValueError) as err:
            raise ApiException(f"Server return malformed response: {result}") from err

    async def post(self, url: str, **kwargs: Any) -> aiohttp.ClientResponse:
        """Make a post request."""
        try:
            resp = await self.request("post", url, **kwargs)
        except ClientError as err:
            raise ApiException(f"Error connecting to API: {err}") from err
        return await AbstractAuth._raise_for_status(resp)

    async def post_json(self, url: str, data_cls: Type[_T], **kwargs: Any) -> _T:
        """Make a post request and return a json response."""
        resp = await self.post(url, **kwargs)
        try:
            result = await resp.text()
        except ClientError as err:
            raise ApiException("Server returned malformed response") from err
        _LOGGER.debug("response=%s", result)
        try:
            return data_cls.from_json(result)
        except (LookupError, ValueError) as err:
            raise ApiException(f"Server return malformed response: {result}") from err

    @classmethod
    async def _raise_for_status(
        cls, resp: aiohttp.ClientResponse
    ) -> aiohttp.ClientResponse:
        """Raise exceptions on failure methods."""
        error_detail = await cls._error_detail(resp)
        try:
            resp.raise_for_status()
        except aiohttp.ClientResponseError as err:
            error_message = f"{err.message} response from API ({resp.status})"
            if error_detail:
                error_message += f": {error_detail}"
            if err.status == HTTPStatus.FORBIDDEN:
                raise ApiForbiddenException(error_message)
            if err.status == HTTPStatus.UNAUTHORIZED:
                raise AuthException(error_message)
            raise ApiException(error_message) from err
        except aiohttp.ClientError as err:
            raise ApiException(f"Error from API: {err}") from err
        return resp

    @classmethod
    async def _error_detail(cls, resp: aiohttp.ClientResponse) -> Error | None:
        """Returns an error message string from the APi response."""
        if resp.status < 400:
            return None
        try:
            result = await resp.text()
        except ClientError:
            return None
        try:
            error_response = ErrorResponse.from_json(result)
        except (LookupError, ValueError):
            return None
        return error_response.error



"""Exceptions for Google Photos api calls."""


class GooglePhotosApiError(Exception):
    """Error talking to the Google Photos API."""


class ApiException(GooglePhotosApiError):
    """Raised during problems talking to the API."""


class AuthException(GooglePhotosApiError):
    """Raised due to auth problems talking to API."""


class ApiForbiddenException(GooglePhotosApiError):
    """Raised due to permission errors talking to API."""


class GooglePhotosParseException(GooglePhotosApiError):
    """Raised when parsing a calendar event fails."""



"""API for Google Photos bound to Home Assistant OAuth.

Callers subclass this to provide an asyncio implementation that handles
refreshing authentication tokens. You then pass in the auth object to the
GooglePhotosLibraryApi object to make authenticated calls to the Google Photos
Library API.

Example usage:
```python
from aiohttp import ClientSession
from google_photos_library_api import api
from google_photos_library_api import auth

class GooglePhotosAuth(auth.AbstractAuth):
    '''Provide OAuth for Google Photos.'''

    async def async_get_access_token(self) -> str:
        # Your auth implementation details are here

# Create a client library
auth = GooglePhotosAuth()
api = api.GooglePhotosLibraryApi(auth)

# Upload content
with open("image.jpg", "rb") as fd:
    upload_result = await api.upload_content(fd.read(), "image/jpeg")

# Create a media item
await api.create_media_items([
    NewMediaItem(SimpleMediaItem(upload_token=upload_result.upload_token))
])

# List all media items created by this application
result = await api.list_media_items()
for item in result.media_items:
    print(item.id)
```

"""

import logging
from typing import Any


from aiohttp.client_exceptions import ClientError

from .exceptions import GooglePhotosApiError
from .auth import AbstractAuth
from .model import (
    MediaItem,
    ListMediaItemResult,
    CreateMediaItemsResult,
    UploadResult,
    NewMediaItem,
    UserInfoResult,
    _ListMediaItemResultModel,
    _ListAlbumResultModel,
    ListAlbumResult,
)

__all__ = [
    "GooglePhotosLibraryApi",
]


_LOGGER = logging.getLogger(__name__)

DEFAULT_PAGE_SIZE = 20

# Only included necessary fields to limit response sizes
GET_MEDIA_ITEM_FIELDS = (
    "id,baseUrl,mimeType,filename,mediaMetadata(width,height,photo,video)"
)
LIST_MEDIA_ITEM_FIELDS = f"nextPageToken,mediaItems({GET_MEDIA_ITEM_FIELDS})"
LIST_ALBUMS_FIELDS = (
    "nextPageToken,albums(id,title,coverPhotoBaseUrl,coverPhotoMediaItemId)"
)
USERINFO_API = "https://www.googleapis.com/oauth2/v1/userinfo"


class GooglePhotosLibraryApi:
    """The Google Photos library api client."""

    def __init__(self, auth: AbstractAuth) -> None:
        """Initialize GooglePhotosLibraryApi."""
        self._auth = auth

    async def get_media_item(
        self, media_item_id: str, fields: str | None = None
    ) -> MediaItem:
        """Get all MediaItem resources."""
        return await self._auth.get_json(
            f"v1/mediaItems/{media_item_id}",
            params={"fields": (fields or GET_MEDIA_ITEM_FIELDS)},
            data_cls=MediaItem,
        )

    async def list_media_items(
        self,
        page_size: int | None = None,
        page_token: str | None = None,
        album_id: str | None = None,
        fields: str | None = None,
    ) -> ListMediaItemResult:
        """Get all MediaItem resources."""

        async def get_next_page(
            next_page_token: str | None,
        ) -> _ListMediaItemResultModel:
            return await self._list_media_items_page(
                page_size=page_size,
                page_token=next_page_token,
                album_id=album_id,
                fields=fields,
            )

        page_result = await get_next_page(None)
        result = ListMediaItemResult(page_result, get_next_page)
        return result

    async def _list_media_items_page(
        self,
        page_size: int | None = None,
        page_token: str | None = None,
        album_id: str | None = None,
        fields: str | None = None,
    ) -> _ListMediaItemResultModel:
        """Get all MediaItem resources."""
        args: dict[str, Any] = {
            "pageSize": (page_size or DEFAULT_PAGE_SIZE),
        }
        if page_token is not None:
            args["pageToken"] = page_token
        if album_id is not None:
            if album_id is not None:
                args["albumId"] = album_id
            return await self._auth.post_json(
                "v1/mediaItems:search",
                params={"fields": (fields or LIST_MEDIA_ITEM_FIELDS)},
                json=args,
                data_cls=_ListMediaItemResultModel,
            )
        return await self._auth.get_json(
            "v1/mediaItems",
            params={**args, "fields": (fields or LIST_MEDIA_ITEM_FIELDS)},
            data_cls=_ListMediaItemResultModel,
        )

    async def list_albums(
        self,
        page_size: int | None = None,
        page_token: str | None = None,
        fields: str | None = None,
    ) -> ListAlbumResult:
        """Get all Album resources."""

        async def get_next_page(
            next_page_token: str | None,
        ) -> _ListAlbumResultModel:
            return await self._list_albums_page(
                page_size=page_size,
                page_token=next_page_token,
                fields=fields,
            )

        page_result = await get_next_page(None)
        result = ListAlbumResult(page_result, get_next_page)
        return result

    async def _list_albums_page(
        self,
        page_size: int | None = None,
        page_token: str | None = None,
        fields: str | None = None,
    ) -> _ListAlbumResultModel:
        """Get all Albums resources."""
        params: dict[str, Any] = {
            "pageSize": (page_size or DEFAULT_PAGE_SIZE),
            "fields": (fields or LIST_ALBUMS_FIELDS),
        }
        if page_token is not None:
            params["pageToken"] = page_token
        return await self._auth.get_json(
            "v1/albums",
            params=params,
            data_cls=_ListAlbumResultModel,
        )

    async def upload_content(self, content: bytes, mime_type: str) -> UploadResult:
        """Upload media content to the API and return an upload token."""
        try:
            result = await self._auth.post(
                "v1/uploads", headers=_upload_headers(mime_type), data=content
            )
            result.raise_for_status()
            return UploadResult(upload_token=await result.text())
        except ClientError as err:
            raise GooglePhotosApiError(f"Failed to upload content: {err}") from err

    async def create_media_items(
        self,
        new_media_items: list[NewMediaItem],
        album_id: str | None = None,
    ) -> CreateMediaItemsResult:
        """Create a batch of media items and return the ids."""
        request: dict[str, Any] = {
            "newMediaItems": [
                new_media_item.to_dict() for new_media_item in new_media_items
            ],
        }
        if album_id is not None:
            request["albumId"] = album_id
        return await self._auth.post_json(
            "v1/mediaItems:batchCreate",
            json=request,
            data_cls=CreateMediaItemsResult,
        )

    async def get_user_info(self) -> UserInfoResult:
        """Get the user profile info.

        This call requires the userinfo.email scope.
        """
        return await self._auth.get_json(USERINFO_API, data_cls=UserInfoResult)


def _upload_headers(mime_type: str) -> dict[str, Any]:
    """Create the upload headers."""
    return {
        "Content-Type": "application/octet-stream",
        "X-Goog-Upload-Content-Type": mime_type,
        "X-Goog-Upload-Protocol": "raw",
    }



"""Tests for Google Photos library API."""

from typing import Any
from unittest.mock import patch
from collections.abc import AsyncGenerator

import pytest
import aiohttp

from google_photos_library_api.api import GooglePhotosLibraryApi
from google_photos_library_api.model import (
    CreateMediaItemsResult,
    NewMediaItemResult,
    UploadResult,
    NewMediaItem,
    SimpleMediaItem,
    Status,
    UserInfoResult,
)
from google_photos_library_api.model import MediaItem, Album


from .conftest import AuthCallback


FAKE_MEDIA_ITEM = {
    "id": "media-item-id-1",
    "description": "Photo 1",
}
FAKE_MEDIA_ITEM2 = {
    "id": "media-item-id-2",
    "description": "Photo 2",
}
FAKE_LIST_MEDIA_ITEMS = {
    "mediaItems": [FAKE_MEDIA_ITEM],
}


@pytest.fixture(name="get_user_info")
async def mock_get_user_info() -> list[dict[str, Any]]:
    """Fixture for returning fake user info responses."""
    return []


@pytest.fixture(name="get_media_items")
async def mock_get_media_item() -> list[dict[str, Any]]:
    """Fixture for fake list media items responses."""
    return []


@pytest.fixture(name="list_media_items")
async def mock_list_media_items() -> list[dict[str, Any]]:
    """Fixture for fake list media items responses."""
    return []


@pytest.fixture(name="search_media_items")
async def mock_search_media_items() -> list[dict[str, Any]]:
    """Fixture for fake search media items responses."""
    return []


@pytest.fixture(name="list_albums")
async def mock_list_albums() -> list[dict[str, Any]]:
    """Fixture for fake list albums responses."""
    return []


@pytest.fixture(name="upload_media_items")
async def mock_upload_media_items() -> list[str]:
    """Fixture for fake list upload endpoint responses."""
    return []


@pytest.fixture(name="create_media_items")
async def mock_create_media_items() -> list[dict[str, Any]]:
    """Fixture for fake create media items responses."""
    return []


@pytest.fixture(name="requests")
async def mock_requests() -> list[aiohttp.web.Request]:
    """Fixture for fake create media items responses."""
    return []


@pytest.fixture(name="api")
async def mock_api(
    auth_cb: AuthCallback,
    requests: list[aiohttp.web.Request],
    get_user_info: list[dict[str, Any]],
    get_media_items: list[dict[str, Any]],
    list_media_items: list[dict[str, Any]],
    search_media_items: list[dict[str, Any]],
    list_albums: list[dict[str, Any]],
    upload_media_items: list[str],
    create_media_items: list[dict[str, Any]],
) -> AsyncGenerator[GooglePhotosLibraryApi, None]:
    """Fixture for fake API object."""

    async def get_user_info_handler(
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        requests.append(request)
        return aiohttp.web.json_response(get_user_info.pop(0))

    async def get_media_items_handler(
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        requests.append(request)
        return aiohttp.web.json_response(get_media_items.pop(0))

    async def list_media_items_handler(
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        requests.append(request)
        return aiohttp.web.json_response(list_media_items.pop(0))

    async def search_media_items_handler(
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        requests.append(request)
        return aiohttp.web.json_response(search_media_items.pop(0))

    async def list_albums_handler(
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        requests.append(request)
        return aiohttp.web.json_response(list_albums.pop(0))

    async def upload_media_items_handler(
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        requests.append(request)
        return aiohttp.web.Response(body=upload_media_items.pop(0))

    async def create_media_items_handler(
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        requests.append(request)
        return aiohttp.web.json_response(create_media_items.pop(0))

    with patch("google_photos_library_api.api.USERINFO_API", "v1/userInfo"):
        auth = await auth_cb(
            [
                ("/v1/userInfo", get_user_info_handler),
                ("/v1/mediaItems", list_media_items_handler),
                ("/v1/mediaItems/{media_item_id}", get_media_items_handler),
                ("/v1/mediaItems:search", search_media_items_handler),
                ("/v1/albums", list_albums_handler),
                ("/v1/uploads", upload_media_items_handler),
                ("/v1/mediaItems:batchCreate", create_media_items_handler),
            ]
        )
        yield GooglePhotosLibraryApi(auth)


async def test_get_user_info(
    api: GooglePhotosLibraryApi,
    get_user_info: list[dict[str, Any]],
    requests: list[aiohttp.web.Request],
) -> None:
    """Test get user info API."""

    get_user_info.append(
        {
            "id": "user-id-1",
            "name": "User Name",
            "given_name": "User Given Name",
            "family_name": "User Full Name",
            "picture": "http://example.com/profile.jpg",
        }
    )
    result = await api.get_user_info()
    assert result == UserInfoResult(
        id="user-id-1",
        name="User Name",
    )


@pytest.mark.parametrize(
    ("list_media_items_requests", "expected_result", "expected_page_token"),
    [
        (
            [FAKE_LIST_MEDIA_ITEMS],
            [MediaItem(id="media-item-id-1", description="Photo 1")],
            None,
        ),
        ([{}], [], None),
        ([{"nextPageToken": "example-token-1"}], [], "example-token-1"),
    ],
)
async def test_list_media_items(
    api: GooglePhotosLibraryApi,
    list_media_items: list[dict[str, Any]],
    requests: list[aiohttp.web.Request],
    list_media_items_requests: list[dict[str, Any]],
    expected_result: list[MediaItem],
    expected_page_token: str | None,
) -> None:
    """Test list media_items API."""

    list_media_items.extend(list_media_items_requests)
    result = await api.list_media_items()
    assert result.media_items == expected_result
    assert result.next_page_token == expected_page_token
    assert len(requests) == 1
    assert requests[0].method == "GET"
    assert requests[0].path == "/path-prefix/v1/mediaItems"
    assert (
        requests[0].query_string
        == "pageSize=20&fields=nextPageToken,mediaItems(id,baseUrl,mimeType,filename,mediaMetadata(width,height,photo,video))"
    )


async def test_list_items_in_album(
    api: GooglePhotosLibraryApi,
    search_media_items: list[dict[str, Any]],
    requests: list[aiohttp.web.Request],
) -> None:
    """Test list media_items API limited to a specific album."""

    search_media_items.append(FAKE_LIST_MEDIA_ITEMS)
    result = await api.list_media_items(album_id="album-id-1")
    assert result.media_items == [
        MediaItem(id="media-item-id-1", description="Photo 1")
    ]
    assert len(requests) == 1
    assert requests[0].method == "POST"
    assert requests[0].path == "/path-prefix/v1/mediaItems:search"
    assert (
        requests[0].query_string
        == "fields=nextPageToken,mediaItems(id,baseUrl,mimeType,filename,mediaMetadata(width,height,photo,video))"
    )


@pytest.mark.parametrize(
    ("fields", "expected_fields"),
    [
        (
            None,
            "nextPageToken,mediaItems(id,baseUrl,mimeType,filename,mediaMetadata(width,height,photo,video))",
        ),
        (
            "nextPageToken,mediaItems(id,description)",
            "nextPageToken,mediaItems(id,description)",
        ),
    ],
    ids=("default_fields", "custom_fields"),
)
async def test_list_media_items_paging(
    api: GooglePhotosLibraryApi,
    list_media_items: list[dict[str, Any]],
    requests: list[aiohttp.web.Request],
    fields: str | None,
    expected_fields: str,
) -> None:
    """Test list media_items API."""

    list_media_items.append(
        {
            "mediaItems": [FAKE_MEDIA_ITEM],
            "nextPageToken": "next-page-token-1",
        }
    )
    list_media_items.append(
        {
            "mediaItems": [FAKE_MEDIA_ITEM2],
        }
    )
    result = await api.list_media_items(fields=fields)
    media_items = []
    async for result_page in result:
        media_items.extend(result_page.media_items)
    assert media_items == [
        MediaItem(id="media-item-id-1", description="Photo 1"),
        MediaItem(id="media-item-id-2", description="Photo 2"),
    ]
    assert len(requests) == 2
    assert requests[0].method == "GET"
    assert requests[0].path == "/path-prefix/v1/mediaItems"
    assert requests[0].query_string == f"pageSize=20&fields={expected_fields}"
    assert requests[1].method == "GET"
    assert requests[1].path == "/path-prefix/v1/mediaItems"
    assert (
        requests[1].query_string
        == f"pageSize=20&pageToken=next-page-token-1&fields={expected_fields}"
    )


@pytest.mark.parametrize(
    "list_args",
    [
        {"album_id": "album-id-1"},
    ],
)
@pytest.mark.parametrize(
    ("fields", "expected_fields"),
    [
        (
            None,
            "nextPageToken,mediaItems(id,baseUrl,mimeType,filename,mediaMetadata(width,height,photo,video))",
        ),
        (
            "nextPageToken,mediaItems(id,description)",
            "nextPageToken,mediaItems(id,description)",
        ),
    ],
    ids=("default_fields", "custom_fields"),
)
async def test_search_items_paging(
    api: GooglePhotosLibraryApi,
    search_media_items: list[dict[str, Any]],
    requests: list[aiohttp.web.Request],
    list_args: dict[str, Any],
    fields: str | None,
    expected_fields: str | None,
) -> None:
    """Test list media_items API."""

    search_media_items.append(
        {
            "mediaItems": [FAKE_MEDIA_ITEM],
            "nextPageToken": "next-page-token-1",
        }
    )
    search_media_items.append(
        {
            "mediaItems": [FAKE_MEDIA_ITEM2],
        }
    )
    result = await api.list_media_items(**list_args, fields=fields)
    media_items = []
    async for result_page in result:
        media_items.extend(result_page.media_items)
    assert media_items == [
        MediaItem(id="media-item-id-1", description="Photo 1"),
        MediaItem(id="media-item-id-2", description="Photo 2"),
    ]
    assert len(requests) == 2
    assert requests[0].method == "POST"
    assert requests[0].path == "/path-prefix/v1/mediaItems:search"
    assert requests[0].query_string == f"fields={expected_fields}"
    assert requests[1].method == "POST"
    assert requests[1].path == "/path-prefix/v1/mediaItems:search"
    assert requests[1].query_string == f"fields={expected_fields}"


@pytest.mark.parametrize(
    ("fields", "expected_fields"),
    [
        (None, "id,baseUrl,mimeType,filename,mediaMetadata(width,height,photo,video)"),
        ("id,description", "id,description"),
    ],
    ids=("default_fields", "custom_fields"),
)
async def test_get_media_item(
    api: GooglePhotosLibraryApi,
    get_media_items: list[dict[str, Any]],
    requests: list[aiohttp.web.Request],
    fields: str | None,
    expected_fields: str | None,
) -> None:
    """Test get media_items API."""

    get_media_items.append(FAKE_MEDIA_ITEM)
    result = await api.get_media_item("media-item-id-1", fields=fields)
    assert result == MediaItem(id="media-item-id-1", description="Photo 1")

    assert len(requests) == 1
    assert requests[0].method == "GET"
    assert requests[0].path == "/path-prefix/v1/mediaItems/media-item-id-1"
    assert requests[0].query_string == f"fields={expected_fields}"


@pytest.mark.parametrize(
    ("fields", "expected_fields"),
    [
        (
            None,
            "nextPageToken,albums(id,title,coverPhotoBaseUrl,coverPhotoMediaItemId)",
        ),
        ("nextPageToken,albums(id,title)", "nextPageToken,albums(id,title)"),
    ],
    ids=("default_fields", "custom_fields"),
)
async def test_list_albums(
    api: GooglePhotosLibraryApi,
    list_albums: list[dict[str, Any]],
    requests: list[aiohttp.web.Request],
    fields: str | None,
    expected_fields: str,
) -> None:
    """Test list albums API."""

    list_albums.append(
        {
            "albums": [
                {
                    "id": "album-id-1",
                    "title": "Album 1",
                    "productUrl": "http://photos.google.com/album/album-id-1",
                }
            ],
            "nextPageToken": "next-page-token-1",
        }
    )
    list_albums.append(
        {
            "albums": [
                {
                    "id": "album-id-2",
                    "title": "Album 2",
                    "productUrl": "http://photos.google.com/album/album-id-2",
                }
            ],
        }
    )
    result = await api.list_albums(fields=fields)
    albums = []
    async for result_page in result:
        albums.extend(result_page.albums)
    assert albums == [
        Album(
            id="album-id-1",
            title="Album 1",
            product_url="http://photos.google.com/album/album-id-1",
        ),
        Album(
            id="album-id-2",
            title="Album 2",
            product_url="http://photos.google.com/album/album-id-2",
        ),
    ]

    assert len(requests) == 2
    assert requests[0].method == "GET"
    assert requests[0].path == "/path-prefix/v1/albums"
    assert requests[0].query_string == f"pageSize=20&fields={expected_fields}"
    assert requests[1].method == "GET"
    assert requests[1].path == "/path-prefix/v1/albums"
    assert (
        requests[1].query_string
        == f"pageSize=20&fields={expected_fields}&pageToken=next-page-token-1"
    )


async def test_upload_items(
    api: GooglePhotosLibraryApi, upload_media_items: list[str]
) -> None:
    """Test list upload_items API."""

    upload_media_items.append("fake-upload-token-1")
    result = await api.upload_content(b"content", "image/jpeg")
    assert result == UploadResult(upload_token="fake-upload-token-1")


@pytest.mark.parametrize(
    "status",
    [
        {
            "code": 200,
            "message": "Success",
        },
        {
            "code": 200,
        },
    ],
)
async def test_create_media_items(
    api: GooglePhotosLibraryApi,
    create_media_items: list[dict[str, Any]],
    status: dict[str, Any],
) -> None:
    """Test list create media items API."""

    create_media_items.append(
        {
            "newMediaItemResults": [
                {
                    "uploadToken": "new-upload-token-1",
                    "status": status,
                    "mediaItem": FAKE_MEDIA_ITEM,
                }
            ]
        }
    )
    result = await api.create_media_items(
        [NewMediaItem(SimpleMediaItem(upload_token="new-upload-token-1"))]
    )
    assert result == CreateMediaItemsResult(
        new_media_item_results=[
            NewMediaItemResult(
                upload_token="new-upload-token-1",
                status=Status(**status),
                media_item=MediaItem(id="media-item-id-1", description="Photo 1"),
            )
        ]
    )



"""Tests for the request client library."""

import aiohttp
import re
import pytest
from dataclasses import dataclass, field

from mashumaro.mixins.json import DataClassJSONMixin
from mashumaro import field_options

from google_photos_library_api.auth import AbstractAuth
from google_photos_library_api.exceptions import (
    ApiException,
    ApiForbiddenException,
)

from .conftest import AuthCallback


@dataclass
class Response(DataClassJSONMixin):
    """Response from listing media items."""

    some_key: str = field(metadata=field_options(alias="some-key"))


class FakeAuth(AbstractAuth):
    """Implementation of AbstractAuth for use in tests."""

    async def async_get_access_token(self) -> str:
        """Return an OAuth credential for the calendar API."""
        return "some-token"


async def test_get_response(auth_cb: AuthCallback) -> None:
    """Test post that returns json."""

    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        body = await request.json()
        assert body == {"client_id": "some-client-id"}
        return aiohttp.web.json_response(
            {
                "some-key": "some-value",
            }
        )

    auth = await auth_cb([("/some-path", handler)])
    data = await auth.get("some-path", json={"client_id": "some-client-id"})
    assert await data.json() == {"some-key": "some-value"}


async def test_get_json_response_unexpected(auth_cb: AuthCallback) -> None:
    """Test json response with wrong response type."""

    async def handler(_: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.json_response(["value1", "value2"])

    @dataclass
    class Response(DataClassJSONMixin):
        """Response from listing media items."""

        items: list[str]

    auth = await auth_cb([("/some-path", handler)])
    with pytest.raises(ApiException):
        await auth.get_json("some-path", data_cls=Response)


async def test_get_json_response(auth_cb: AuthCallback) -> None:
    """Test post that returns json."""

    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        body = await request.json()
        assert body == {"client_id": "some-client-id"}
        return aiohttp.web.json_response(
            {
                "some-key": "some-value",
            }
        )

    auth = await auth_cb([("/some-path", handler)])
    data = await auth.get_json(
        "some-path", json={"client_id": "some-client-id"}, data_cls=Response
    )
    assert data == Response(some_key="some-value")


async def test_post_json_response(auth_cb: AuthCallback) -> None:
    """Test post that returns json."""

    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        body = await request.json()
        assert body == {"client_id": "some-client-id"}
        return aiohttp.web.json_response(
            {
                "some-key": "some-value",
            }
        )

    auth = await auth_cb([("/some-path", handler)])
    data = await auth.post_json(
        "some-path", json={"client_id": "some-client-id"}, data_cls=Response
    )
    assert data == Response(some_key="some-value")


async def test_post_json_response_unexpected(auth_cb: AuthCallback) -> None:
    """Test post that returns wrong json type."""

    async def handler(_: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.json_response(["value1", "value2"])

    auth = await auth_cb([("/some-path", handler)])

    with pytest.raises(ApiException):
        await auth.post_json("some-path", data_cls=Response)


async def test_post_json_response_unexpected_text(auth_cb: AuthCallback) -> None:
    """Test post that returns unexpected format."""

    async def handler(_: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.Response(text="body")

    auth = await auth_cb([("/some-path", handler)])

    with pytest.raises(ApiException):
        await auth.post_json("some-path", data_cls=Response)


async def test_get_json_response_bad_request(auth_cb: AuthCallback) -> None:
    """Test error handling with detailed json response."""

    async def handler(_: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.json_response(
            {
                "error": {
                    "errors": [
                        {
                            "domain": "calendar",
                            "reason": "timeRangeEmpty",
                            "message": "The specified time range is empty.",
                            "locationType": "parameter",
                            "location": "timeMax",
                        }
                    ],
                    "code": 400,
                    "message": "The specified time range is empty.",
                }
            },
            status=400,
        )

    auth = await auth_cb([("/some-path", handler)])

    with pytest.raises(
        ApiException,
        match=re.escape(
            "Bad Request response from API (400): 400: The specified time range is empty."
        ),
    ):
        await auth.get("some-path")

    with pytest.raises(
        ApiException,
        match=re.escape(
            "Bad Request response from API (400): 400: The specified time range is empty."
        ),
    ):
        await auth.get_json("some-path", data_cls=Response)

    with pytest.raises(
        ApiException,
        match=re.escape(
            "Bad Request response from API (400): 400: The specified time range is empty."
        ),
    ):
        await auth.post("some-path")

    with pytest.raises(
        ApiException,
        match=re.escape(
            "Bad Request response from API (400): 400: The specified time range is empty."
        ),
    ):
        await auth.post_json("some-path", data_cls=Response)


async def test_unavailable_error(auth_cb: AuthCallback) -> None:
    """Test of basic request/response handling."""

    async def handler(_: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.Response(status=500)

    auth = await auth_cb([("/some-path", handler)])

    with pytest.raises(ApiException):
        await auth.get_json("some-path", data_cls=Response)


async def test_forbidden_error(auth_cb: AuthCallback) -> None:
    """Test request/response handling for 403 status."""

    async def handler(_: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.json_response(
            {
                "error": {
                    "code": 403,
                    "message": "Google Photos API has not been used in project 0 before or it is disabled. Enable it by visiting https://console.developers.google.com/apis/library/photoslibrary.googleapis.com/overview?project=0 then retry. If you enabled this API recently, wait a few minutes for the action to propagate to our systems and retry.",
                    "status": "PERMISSION_DENIED",
                }
            },
            status=403,
        )

    auth = await auth_cb([("/some-path", handler)])

    with pytest.raises(
        ApiForbiddenException,
        match=re.escape(
            "Forbidden response from API (403): PERMISSION_DENIED (403): Google Photos API has not been used in project 0 before or it is disabled. Enable it by visiting https://console.developers.google.com/apis/library/photoslibrary.googleapis.com/overview?project=0 then retry. If you enabled this API recently, wait a few minutes for the action to propagate to our systems and retry."
        ),
    ):
        await auth.get_json("some-path", data_cls=Response)


async def test_error_detail_parse_error(auth_cb: AuthCallback) -> None:
    """Test request/response handling for 403 status."""

    async def handler(_: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.Response(status=403, text="Plain text error message")

    auth = await auth_cb([("/some-path", handler)])

    with pytest.raises(
        ApiForbiddenException, match=re.escape("Forbidden response from API (403)")
    ):
        await auth.get_json("some-path", data_cls=Response)


async def test_invalid_argument(auth_cb: AuthCallback) -> None:
    """Test request/response handling for 403 status."""

    async def handler(_: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.json_response(
            {
                "error": {
                    "code": 400,
                    "message": "Request contains an invalid argument.",
                    "status": "INVALID_ARGUMENT",
                    "details": [
                        {
                            "@type": "type.googleapis.com/google.rpc.BadRequest",
                            "fieldViolations": [
                                {
                                    "field": "id,baseUrl,mimeType,filename,mediaMetadata(width,height,photo,video)",
                                    "description": "Error expanding 'fields' parameter. Cannot find matching fields for path 'id'."
                                }
                            ]
                        }
                    ]
                }
            },
            status=400,
        )

    auth = await auth_cb([("/some-path", handler)])

    with pytest.raises(
        ApiException,
        match=re.escape(
            "Bad Request response from API (400): INVALID_ARGUMENT (400): Request contains an invalid argument."
        ) + "\nError details: .*",
    ):
        await auth.get_json("some-path", data_cls=Response)



"""Google Photos Library API Data Model."""

from collections.abc import AsyncIterator, Awaitable, Callable
from dataclasses import dataclass, field
from typing import Any, Self

from mashumaro import DataClassDictMixin, field_options
from mashumaro.mixins.json import DataClassJSONMixin

__all__ = [
    "ListMediaItemResult",
    "MediaItem",
    "MediaMetadata",
    "Photo",
    "Video",
    "ContributorInfo",
    "UploadResult",
    "NewMediaItem",
    "SimpleMediaItem",
    "CreateMediaItemsResult",
    "NewMediaItemResult",
    "Status",
    "UserInfoResult",
]


@dataclass
class Photo(DataClassDictMixin):
    """Metadata for a photo media item."""

    camera_make: str | None = field(
        metadata=field_options(alias="cameraMake"), default=None
    )
    """Make of the camera that took the photo."""

    camera_model: str | None = field(
        metadata=field_options(alias="cameraModel"), default=None
    )
    """Model of the camera that took the photo."""

    focal_length: float | None = field(
        metadata=field_options(alias="focalLength"), default=None
    )
    """Focal length of the camera lens used to take the photo."""

    aperture_f_number: float | None = field(
        metadata=field_options(alias="apertureFNumber"), default=None
    )
    """Aperture f number of the camera lens used to take the photo."""

    iso_equivalent: int | None = field(
        metadata=field_options(alias="isoEquivalent"), default=None
    )
    """ISO value that the camera used to take the photo."""

    exposure_time: str | None = field(
        metadata=field_options(alias="exposureTime"), default=None
    )
    """Exposure time (duraton like '3.5s') of the camera lens aperture when the photo was taken."""


@dataclass
class Video(DataClassDictMixin):
    """Metadata for a video media item."""

    camera_make: str | None = field(
        metadata=field_options(alias="cameraMake"), default=None
    )
    """Make of the camera that took the video."""

    camera_model: str | None = field(
        metadata=field_options(alias="cameraModel"), default=None
    )
    """Model of the camera that took the video."""

    fps: str | None = None
    """Frames per second of the video."""

    status: str | None = None
    """Status of the video."""


@dataclass
class MediaMetadata(DataClassDictMixin):
    """Metadata for a media item."""

    creation_time: str | None = field(
        metadata=field_options(alias="creationTime"), default=None
    )
    """Creation time of the media item."""

    width: int | None = None
    """Width of the media item in pixels."""

    height: int | None = None
    """Height of the media item in pixels."""

    photo: Photo | None = None
    """Metadata for a photo media item."""

    video: Video | None = None
    """Metadata for a video media item."""


@dataclass
class ContributorInfo(DataClassDictMixin):
    """Information about the user who contributed this media item."""

    profile_picture_base_url: str = field(
        metadata=field_options(alias="profilePictureBaseUrl")
    )
    """Base URL for the user's profile picture."""

    display_name: str = field(metadata=field_options(alias="displayName"))
    """Display name of the user."""


@dataclass
class MediaItem(DataClassJSONMixin):
    """Representation of a media item (such as a photo or video) in Google Photos."""

    id: str
    """Identifier for the media item, used between sessions to identify this media item."""

    description: str | None = None
    """Description of the media item, shown to the user in the item's info section in the Google Photos app."""

    product_url: str | None = field(
        metadata=field_options(alias="product_url"), default=None
    )
    """Google Photos URL for the media item."""

    base_url: str | None = field(metadata=field_options(alias="baseUrl"), default=None)
    """A URL to the media item's bytes."""

    mime_type: str | None = field(
        metadata=field_options(alias="mimeType"), default=None
    )
    """MIME type of the media item."""

    media_metadata: MediaMetadata | None = field(
        metadata=field_options(alias="mediaMetadata"), default=None
    )
    """Metadata related to the media item, such as, height, width, or creation time."""

    contributor_info: ContributorInfo | None = field(
        metadata=field_options(alias="contributorInfo"), default=None
    )
    """Information about the user who created the media item."""

    filename: str | None = None
    """Filename of the media item."""


@dataclass
class Album(DataClassDictMixin):
    """Representation of an album in Google Photos."""

    id: str
    """Identifier for the album, used between sessions to identify this album."""

    title: str
    """Title of the album."""

    product_url: str | None = field(metadata=field_options(alias="productUrl"), default=None)
    """Google Photos URL for the album."""

    media_items_count: int | None = field(
        metadata=field_options(alias="mediaItemsCount"), default=None
    )
    """Number of media items in the album."""

    cover_photo_base_url: str | None = field(
        metadata=field_options(alias="coverPhotoBaseUrl"), default=None
    )
    """Base URL for the cover photo of the album."""

    cover_photo_media_item_id: str | None = field(
        metadata=field_options(alias="coverPhotoMediaItemId"),
        default=None,
    )
    """Identifier for the cover photo of the album."""

    is_writeable: bool | None = field(
        metadata=field_options(alias="isWriteable"), default=None
    )
    """Whether the album is writable."""


@dataclass
class UserInfoResult(DataClassJSONMixin):
    """Response from getting user info."""

    id: str
    """User ID."""

    name: str
    """User name."""


@dataclass
class _ListMediaItemResultModel(DataClassJSONMixin):
    """Api response containing a list of events."""

    media_items: list[MediaItem] = field(metadata=field_options(alias="mediaItems"), default_factory=list)
    """List of media items."""

    next_page_token: str | None = field(
        metadata=field_options(alias="nextPageToken"), default=None
    )
    """Token for the next page of results."""


class ListMediaItemResult:
    """Response from listing media items."""

    def __init__(
        self,
        response: _ListMediaItemResultModel,
        get_next_page: (
            Callable[[str | None], Awaitable[_ListMediaItemResultModel]] | None
        ) = None,
    ):
        self._response = response
        self._get_next_page = get_next_page

    @property
    def media_items(self) -> list[MediaItem]:
        """List of media items."""
        return self._response.media_items

    @property
    def next_page_token(self) -> str | None:
        """Token for the next page of results."""
        return self._response.next_page_token

    async def __aiter__(self) -> AsyncIterator[Self]:
        """Async iterator to traverse through pages of responses."""
        response = self
        while response is not None:
            yield response
            if not response.next_page_token or not self._get_next_page:
                break
            page_result = await self._get_next_page(response.next_page_token)
            response = self.__class__(page_result)


@dataclass
class _ListAlbumResultModel(DataClassJSONMixin):
    """Api response containing a list of albums requested."""

    albums: list[Album] = field(default_factory=list)
    """List of albums shown in the Albums tab of the user's Google Photos app."""

    next_page_token: str | None = field(
        metadata=field_options(alias="nextPageToken"), default=None
    )
    """Token to use to get the next set of albums."""


class ListAlbumResult:
    """Api response containing a list of albums requested."""

    def __init__(
        self,
        response: _ListAlbumResultModel,
        get_next_page: (
            Callable[[str | None], Awaitable[_ListAlbumResultModel]] | None
        ) = None,
    ):
        self._response = response
        self._get_next_page = get_next_page

    @property
    def albums(self) -> list[Album]:
        """List of albums shown in the Albums tab of the user's Google Photos app."""
        return self._response.albums

    @property
    def next_page_token(self) -> str | None:
        """Token to use to get the next set of albums."""
        return self._response.next_page_token

    async def __aiter__(self) -> AsyncIterator[Self]:
        """Async iterator to traverse through pages of responses."""
        response = self
        while response is not None:
            yield response
            if not response.next_page_token or not self._get_next_page:
                break
            page_result = await self._get_next_page(response.next_page_token)
            response = self.__class__(page_result)


@dataclass
class Status(DataClassJSONMixin):
    """Status of the media item."""

    code: int
    """The status code, which should be an enum value of google.rpc.Code"""

    message: str | None = None
    """A developer-facing error message, which should be in English"""

    details: list[dict[str, Any]] = field(default_factory=list)
    """A list of messages that carry the error details"""


@dataclass
class SimpleMediaItem:
    """Simple media item."""

    upload_token: str = field(metadata=field_options(alias="uploadToken"))
    """Upload token for the media item."""

    file_name: str | None = field(
        metadata=field_options(alias="fileName"), default=None
    )
    """Filename of the media item."""


@dataclass
class NewMediaItem(DataClassDictMixin):
    """New media item to create."""

    simple_media_item: SimpleMediaItem = field(
        metadata=field_options(alias="simpleMediaItem")
    )
    """Simple media item to create."""

    description: str | None = None
    """Description of the media item."""


@dataclass
class NewMediaItemResult(DataClassJSONMixin):
    """Result of creating a new media item."""

    upload_token: str = field(metadata=field_options(alias="uploadToken"))
    """Upload token for the media item."""

    status: Status | None = None
    """If an error occurred during the creation of this media item, this field is populated with information related to the error."""

    media_item: MediaItem | None = field(
        metadata=field_options(alias="mediaItem"), default=None
    )
    """Media item created with the upload token."""


@dataclass
class UploadResult:
    """Response from uploading media items."""

    upload_token: str
    """Upload token for the media item."""


@dataclass
class CreateMediaItemsResult(DataClassJSONMixin):
    """Response from creating media items."""

    new_media_item_results: list[NewMediaItemResult] = field(
        metadata=field_options(alias="newMediaItemResults")
    )
    """List of created media items."""


@dataclass
class Error:
    """Error details from the API response."""

    status: str | None = None
    code: int | None = None
    message: str | None = None
    details: list[dict[str, Any]] | None = field(default_factory=list)

    def __str__(self) -> str:
        """Return a string representation of the error details."""
        error_message = ""
        if self.status:
            error_message += self.status
        if self.code:
            if error_message:
                error_message += f" ({self.code})"
            else:
                error_message += str(self.code)
        if self.message:
            if error_message:
                error_message += ": "
            error_message += self.message
        if self.details:
            error_message += f"\nError details: ({self.details})"
        return error_message


@dataclass
class ErrorResponse(DataClassJSONMixin):
    """A response message that contains an error message."""

    error: Error | None = None



"""Constants for Google Photos Library API."""

LIBRARY_API_URL = "https://photoslibrary.googleapis.com"
USERINFO_API = "https://www.googleapis.com/oauth2/v1/userinfo"


class LibraryApiScope:
    """Google Photos OAuth scopes.

    Note that other scopes that access the entire library are deprecated.
    """

    APPEND_ONLY = "https://www.googleapis.com/auth/photoslibrary.appendonly"
    READONLY_APP_CREATED_DATA = (
        "https://www.googleapis.com/auth/photoslibrary.readonly.appcreateddata"
    )
    EDIT_APP_CREATED_DATA = (
        "https://www.googleapis.com/auth/photoslibrary.edit.appcreateddata"
    )
